from hiword import extract_keywords

docs = [
    """新型冠状病毒肺炎（Corona Virus Disease 2019，COVID-19），简称“新冠肺炎”，世界卫生组织命名为“2019冠状病毒病” [1-2]  ，是指2019新型冠状病毒感染导致的肺炎。2019年12月以来，湖北省武汉市部分医院陆续发现了多例有华南海鲜市场暴露史的不明原因肺炎病例，证实为2019新型冠状病毒感染引起的急性呼吸道传染病。 [3-6] 2020年2月11日，世界卫生组织总干事谭德塞在瑞士日内瓦宣布，将新型冠状病毒感染的肺炎命名为“COVID-19” [7]  。2月22日，国家卫生健康委发布通知，“新型冠状病毒肺炎”英文名称修订为“COVID-19” [8]  。3月11日，世卫组织认为当前新冠肺炎疫情可被称为全球大流行 [9]  。4月4日，中国举行全国性哀悼活动 [10]  。8月18日，国家卫健委修订完成了新型冠状病毒肺炎诊疗方案（试行第八版） [11]  。9月3日，世卫组织新冠肺炎疫情应对评估专家组共同主席在成员国吹风会上宣布了专家组成员名单，钟南山入选。 [12] 截至2021年2月6日24时，据31个省（自治区、直辖市）和新疆生产建设兵团报告，现有确诊病例1176例（其中重症病例25例），累计治愈出院病例83880例，累计死亡病例4636例，累计报告确诊病例89692例，现有疑似病例3例。累计追踪到密切接触者970667人，尚在医学观察的密切接触者30212人。 [13]""",
]


def test_extract_keywords():
    for i, s in enumerate(docs):
        print('#{}: {}'.format(i + 1, s))
        print('#keywords:')
        for x, w in extract_keywords(s, with_weight=True):
            print(x, w)



import math
import re
from collections import defaultdict
from typing import Dict, List, Union

from jieba import Tokenizer

from hiword.dataloader import DictLoader, IDFLoader, StopwordsLoader
from hiword.filter import WordFilterChain, NumericFilter
from hiword.utils import traditional_to_simple


class KeywordsExtractor:
    def __init__(self):
        self.tokenizer = Tokenizer()
        self.dict = DictLoader()
        self.idf = IDFLoader()
        self.stopwords = StopwordsLoader()

        self.word_filter = WordFilterChain(
            NumericFilter(),
        )

    def extract_keywords(self, doc):
        words = self._tokenize(doc)
        keywords = self._extract_shot_keywords(words)
        long_words = self._detect_long_keywords(words, keywords)

        freq = defaultdict(lambda: 0)
        for word, count in keywords:
            if self.word_filter.filter(word):
                continue
            freq[word] += count
        for word, count, _ in long_words:
            if self.word_filter.filter(word):
                continue
            freq[word] += count

        res = self._rerank_words(freq)
        res = self._filter_included_words(res, freq)
        return res

    def _tokenize(self, doc: Union[str, List[str]]) -> List[str]:
        if isinstance(doc, str):
            words = []
            for i in self.tokenizer.cut(doc, HMM=False):
                w = i.strip()
                if len(w) == 0:
                    continue
                words.append(w)
        else:
            words = doc
        return words

    def _filter_included_words(self, words: list, freq: Dict[str, int]):
        not_merge = words
        merged = []
        for i in range(len(not_merge)):
            w1 = not_merge[i][0]
            w1_ok = True
            for j in range(len(not_merge)):
                w2 = not_merge[j][0]
                if len(w1) < len(w2) and w1 in w2 and freq[w1] == freq[w2]:
                    w1_ok = False
                    break
            if w1_ok:
                merged.append(not_merge[i])
        return merged

    def _detect_long_keywords(self, words, keywords):
        keywords = set([t[0] for t in keywords])
        appears = defaultdict(lambda: 0)
        parts = defaultdict(list)

        def commit(continuous_words):
            end_pos = len(continuous_words)
            if end_pos > 0:
                for i in range(0, end_pos - 1):
                    for j in range(i + 2, end_pos + 1):
                        word = ''.join(continuous_words[i:j])
                        appears[word] += 1
                        parts[word] = continuous_words[i:j]

        w = []
        for cur in range(0, len(words)):
            x = words[cur]
            if x in keywords:
                w.append(x)
            elif w:
                commit(w)
                w = []
        if w:
            commit(w)

        long_words = []
        for word, count in appears.items():
            # TODO: 合理计算阈值
            if count >= 2:
                long_words.append((word, count, parts[word]))
        return long_words

    def _extract_shot_keywords(self, words) -> list:
        freq = defaultdict(lambda: 0)
        for w in words:
            if self.stopwords.is_stopword(w):
                continue
            freq[w] += 1
        tfidf = {}
        for k in freq:
            tfidf[k] = freq[k] * self.idf.word_idf(k)
        keywords = sorted(tfidf, key=tfidf.__getitem__, reverse=True)
        keywords = [(k, freq[k]) for k in keywords if freq[k] > 1]
        topk = 100
        # TODO: 根据文本长度等信息动态调整简单关键词的数量
        return keywords[:topk]

    def _rerank_words(self, freq: Dict[str, int]) -> list:
        res = []
        for k in freq:
            s = math.log2(freq[k] + 0.5) * (freq[k] / (freq[k] + self.dict.word_freq(k)))
            # TODO: 融合词汇本身的信息量
            res.append((k, s))
        res.sort(key=lambda x: x[1], reverse=True)
        return res


_keywords_extractor = None

_illegal = re.compile(r'[^\u4e00-\u9fa5a-zA-Z0-9}]')


def extract_keywords(doc, with_weight=False):
    global _keywords_extractor
    if _keywords_extractor is None:
        _keywords_extractor = KeywordsExtractor()
    # 繁体转简体
    doc = traditional_to_simple(doc)
    # 过滤无效字符
    doc = _illegal.sub('.', doc)
    keywords = _keywords_extractor.extract_keywords(doc)
    keywords = [k if with_weight else k[0] for k in keywords if k[1] >= 0.1]
    return keywords



from multiprocessing import Lock
from os.path import join, dirname

_zh_hans = {}
_zh_hans_lock = Lock()


def _load():
    file = join(dirname(__file__), 'data', 'zh_hans.txt')
    with open(file) as f:
        while True:
            line = f.readline()
            if not line:
                break
            s = line.split()
            _zh_hans[ord(s[0])] = ord(s[1])


def traditional_to_simple(text: str):
    with _zh_hans_lock:
        if len(_zh_hans) == 0:
            _load()

    return text.translate(_zh_hans)
